from blinker import Signal


class TranslationEvents:
    """Container for blinker signals emitted during translation."""

    preprocessed: Signal
    batch_translated: Signal
    scene_translated: Signal

    def __init__(self):
        self.preprocessed = Signal("translation-preprocessed")
        self.batch_translated = Signal("translation-batch-translated")
        self.scene_translated = Signal("translation-scene-translated")




import logging
from typing import Any
from PySubtrans.Helpers.Localization import _
from PySubtrans.Providers.Clients.OpenAIClient import OpenAIClient
from PySubtrans.SettingsType import SettingsType
from PySubtrans.SubtitleError import TranslationError, TranslationResponseError
from PySubtrans.TranslationPrompt import TranslationPrompt

linesep = '\n'

class OpenAIReasoningClient(OpenAIClient):
    """
    Handles chat communication with OpenAI to request translations using the Responses API
    """
    def __init__(self, settings: SettingsType):
        settings.update({
            'supports_system_messages': True,
            'supports_conversation': True,
            'supports_reasoning': True,
            'supports_system_prompt': True,
            'system_role': 'developer'
        })
        super().__init__(settings)

    @property
    def reasoning_effort(self) -> str:
        return self.settings.get_str( 'reasoning_effort') or "low"
    
    def _send_messages(self, prompt: TranslationPrompt, temperature: float|None) -> dict[str, Any] | None:
        """
        Make a request to OpenAI Responses API for translation
        """
        if not self.client:
            raise TranslationError(_("Client is not initialized"))

        if not self.model:
            raise TranslationError(_("No model specified"))

        if not prompt.content or not isinstance(prompt.content, list):
            raise TranslationError(_("No content provided for translation"))

        result = self.client.responses.create(
            model=self.model,
            input=prompt.content, # type: ignore[arg-type]
            instructions=prompt.system_prompt,
            reasoning={"effort": self.reasoning_effort}  # type: ignore[arg-type]
        )
        
        if self.aborted:
            return None
        
        # Build response with usage info and content
        response = self._extract_usage_info(result)
        text, reasoning = self._extract_text_content(result)
        
        response.update({
            'text': text,
            'finish_reason': self._normalize_finish_reason(result)
        })
        
        if reasoning:
            response['reasoning'] = reasoning
            
        return response
            
    def _extract_text_content(self, result):
        """Extract text content with cleaner fallback logic"""        
        if hasattr(result, 'output') and result.output:
            return self._parse_structured_output(result.output)

        if hasattr(result, 'output_text') and result.output_text:
            return result.output_text, None

        raise TranslationResponseError(_("No text content found in response"), response=result)

    def _parse_structured_output(self, output_blocks):
        """Parse structured output blocks"""
        text_parts = []
        reasoning_parts = []
        
        for block in output_blocks:
            content = getattr(block, 'content', [])
            if content is None:
                continue

            if isinstance(content, str):
                text_parts.append(content)
                continue
                
            for item in content:
                if hasattr(item, 'text') and item.text:
                    if getattr(item, 'type', None) == 'reasoning':
                        reasoning_parts.append(item.text)
                    else:
                        text_parts.append(item.text)
                elif isinstance(item, dict):
                    if item.get('text'):
                        if item.get('type') == 'reasoning':
                            reasoning_parts.append(item['text'])
                        else:
                            text_parts.append(item['text'])

                    # Handle nested reasoning structure
                    if item.get('type') == 'reasoning':
                        r_text = (item.get('reasoning') or {}).get('text')
                        if r_text:
                            reasoning_parts.append(r_text)
        
        return ('\n'.join(text_parts) or None, 
                '\n'.join(reasoning_parts) or None)

    def _extract_usage_info(self, result):
        """Extract token usage information"""
        usage = getattr(result, 'usage', None)
        if not usage:
            return {'response_time': getattr(result, 'response_ms', 0)}
        
        info = {
            'prompt_tokens': getattr(usage, 'input_tokens', None) or getattr(usage, 'prompt_tokens', None),
            'output_tokens': getattr(usage, 'output_tokens', None) or getattr(usage, 'completion_tokens', None),
            'response_time': getattr(result, 'response_ms', 0)
        }
        
        # Calculate total if not provided
        if info['prompt_tokens'] and info['output_tokens']:
            info['total_tokens'] = info['prompt_tokens'] + info['output_tokens']
        
        # Add reasoning-specific tokens
        details = getattr(usage, 'output_tokens_details', None) or getattr(usage, 'completion_tokens_details', None)
        if details:
            reasoning_tokens = getattr(details, 'reasoning_tokens', None)
            accepted_tokens = getattr(details, 'accepted_prediction_tokens', None)
            rejected_tokens = getattr(details, 'rejected_prediction_tokens', None)
            
            if reasoning_tokens is not None:
                info['reasoning_tokens'] = reasoning_tokens
            if accepted_tokens is not None:
                info['accepted_prediction_tokens'] = accepted_tokens
            if rejected_tokens is not None:
                info['rejected_prediction_tokens'] = rejected_tokens
        
        return {k: v for k, v in info.items() if v is not None}

    def _normalize_finish_reason(self, result):
        """Normalize finish reason to legacy format"""
        finish = getattr(result, 'stop_reason', None) or getattr(result, 'finish_reason', None)
        return 'length' if finish == 'max_output_tokens' else finish




import logging
import time
from typing import Any
import httpx

from PySubtrans.Helpers import FormatMessages
from PySubtrans.Helpers.Parse import ParseErrorMessageFromText
from PySubtrans.Helpers.Localization import _
from PySubtrans.Options import SettingsType
from PySubtrans.SubtitleError import TranslationImpossibleError, TranslationResponseError
from PySubtrans.Translation import Translation
from PySubtrans.TranslationClient import TranslationClient
from PySubtrans.TranslationPrompt import TranslationPrompt

class CustomClient(TranslationClient):
    """
    Handles communication with local LLM server to request translations
    """
    def __init__(self, settings : SettingsType):
        super().__init__(settings)
        self.client: httpx.Client|None = None
        self.headers: dict[str, str] = {'Content-Type': 'application/json'}
        self._add_additional_headers(settings)

        if self.api_key:
            self.headers['Authorization'] = f"Bearer {self.api_key}"

        logging.info(_("Translating with server at {server_address}{endpoint}").format(
            server_address=self.server_address, endpoint=self.endpoint
        ))
        if self.model:
            logging.info(_("Using model: {model}").format(model=self.model))

    @property
    def server_address(self) -> str|None:
        return self.settings.get_str( 'server_address')

    @property
    def endpoint(self) -> str|None:
        return self.settings.get_str( 'endpoint')

    @property
    def supports_conversation(self) -> bool:
        return self.settings.get_bool( 'supports_conversation', False)

    @property
    def api_key(self) -> str|None:
        return self.settings.get_str( 'api_key')

    @property
    def model(self) -> str|None:
        return self.settings.get_str( 'model')

    @property
    def max_tokens(self) -> int|None:
        max_tokens = self.settings.get_int( 'max_tokens', 0)
        return max_tokens if max_tokens != 0 else None
    
    @property
    def max_completion_tokens(self) -> int|None:
        max_completion_tokens = self.settings.get_int( 'max_completion_tokens', 0)
        return max_completion_tokens if max_completion_tokens != 0 else None
    
    @property
    def timeout(self) -> int:
        return self.settings.get_int( 'timeout') or 300

    def _request_translation(self, prompt : TranslationPrompt, temperature : float|None = None) -> Translation|None:
        """
        Request a translation based on the provided prompt
        """
        logging.debug(f"Messages:\n{FormatMessages(prompt.messages)}")

        temperature = temperature or self.temperature
        response = self._make_request(prompt, temperature)

        translation = Translation(response) if response else None

        return translation

    def _abort(self) -> None:
        if self.client:
            self.client.close()
        return super()._abort()

    def _make_request(self, prompt : TranslationPrompt, temperature: float|None) -> dict[str, Any]|None:
        """
        Make a request to the server to provide a translation
        """
        response = {}

        for retry in range(self.max_retries + 1):
            if self.aborted:
                return None

            try:
                request_body = self._generate_request_body(prompt, temperature)
                logging.debug(f"Request Body:\n{request_body}")

                if self.server_address is None or self.endpoint is None:
                    raise TranslationImpossibleError(_("Server address or endpoint is not set"))

                self.client = httpx.Client(base_url=self.server_address, follow_redirects=True, timeout=self.timeout, headers=self.headers)

                result : httpx.Response = self.client.post(self.endpoint, json=request_body)

                if self.aborted:
                    return None

                if result.is_error:
                    parsed_message = ParseErrorMessageFromText(result.text)
                    summary_text = parsed_message if parsed_message else result.text
                    if result.is_client_error:
                        raise TranslationResponseError(_("Client error: {status_code} {text}").format(
                            status_code=result.status_code, text=summary_text
                        ), response=result)
                    else:
                        raise TranslationResponseError(_("Server error: {status_code} {text}").format(
                            status_code=result.status_code, text=summary_text
                        ), response=result)

                logging.debug(f"Response:\n{result.text}")

                content = result.json()

                response['model'] = content.get('model')
                response['response_time'] = content.get('response_ms', 0)

                usage = content.get('usage', {})
                response['prompt_tokens'] = usage.get('prompt_tokens')
                response['output_tokens'] = usage.get('completion_tokens')
                response['total_tokens'] = usage.get('total_tokens')
                if 'reasoning_tokens' in usage:
                    response['reasoning_tokens'] = usage.get('reasoning_tokens')

                choices = content.get('choices')
                if not choices:
                    raise TranslationResponseError(_("No choices returned in the response"), response=result)

                for choice in choices:
                    # Try to extract translation from the response choice
                    if 'message' in choice:
                        message = choice.get('message', {})
                        response['finish_reason'] = choice.get('finish_reason')
                        if 'reasoning_content' in message:
                            response['reasoning'] = message['reasoning_content']

                        response['text'] = message.get('content')
                        break

                    if 'text' in choice:
                        response['text'] = choice.get('text')
                        response['finish_reason'] = choice.get('finish_reason')
                        break

                if not response.get('text'):
                    raise TranslationResponseError(_("No text returned in the response"), response=result)

                # Return the response if the API call succeeds
                return response

            except TranslationResponseError:
                raise

            except httpx.ConnectError as e:
                if not self.aborted:
                    logging.error(_("Failed to connect to server at {server_address}{endpoint}").format(
                        server_address=self.server_address, endpoint=self.endpoint
                    ))

            except httpx.NetworkError as e:
                if not self.aborted:
                    logging.error(_("Network error communicating with server: {error}").format(
                        error=str(e)
                    ))

            except httpx.ReadTimeout as e:
                if not self.aborted:
                    logging.error(_("Request to server timed out: {error}").format(
                        error=str(e)
                    ))

            except Exception as e:
                raise TranslationImpossibleError(_("Unexpected error communicating with server"), error=e)

            if self.aborted:
                return None
            
            if retry == self.max_retries:
                raise TranslationImpossibleError(_("Failed to communicate with server after {max_retries} retries").format(
                    max_retries=self.max_retries
                ))

            sleep_time = self.backoff_time * 2.0**retry
            logging.warning(_("Retrying in {sleep_time} seconds...").format(
                sleep_time=sleep_time
            ))
            time.sleep(sleep_time)

    def _generate_request_body(self, prompt: TranslationPrompt, temperature: float|None) -> dict[str, Any]:
        request_body = {
            'temperature': temperature,
            'stream': False
        }

        if self.max_tokens:
            request_body['max_tokens'] = self.max_tokens

        if self.max_completion_tokens:
            request_body['max_completion_tokens'] = self.max_completion_tokens

        if self.model:
            request_body['model'] = self.model

        if self.supports_conversation:
            request_body['messages'] = prompt.messages
        else:
            request_body['prompt'] = prompt.content

        return request_body

    def _add_additional_headers(self, settings):
        additional_headers = settings.get('additional_headers', {})  # Keep dict access for complex types
        if isinstance(additional_headers, dict):
            for key, value in additional_headers.items():
                if isinstance(value, str):
                    self.headers[key] = value

